/* --------------------------------------------------------- */
/*	BASIC-E													 */
/*	Copyright(C)2006 t.hara									 */
/* ========================================================= */
/*	variable manager										 */
/* --------------------------------------------------------- */

#ifndef __VARIABLE_MANAGER_H__
#define __VARIABLE_MANAGER_H__

#include <cstring>
#include "parser.h"
#include "program_list.h"
#include "debug.h"

namespace {
	const int MAX_VARIABLE_LENGTH = 32;
	const int MAX_ARRAY_DIMENTION = 8;
	const int MAX_ELEMENT_MEMBER = 65536;
}

class c_variable {
protected:
	c_variable	*p_prev;
	c_variable	*p_next;

	char	name[ MAX_VARIABLE_LENGTH + 1 ];
	int		reference_counter;
	int		type;

public:
	enum {
		UNKNOWN		= 0,
		I_NUMBER	= 1,	/* integer */
		R_NUMBER	= 2,	/* real */
		STRING		= 3,	/* string */
		I_ARRAY		= 4,	/* integer array */
		R_ARRAY		= 5,	/* real array */
	};

	c_variable( int cnt = 1 ) {
		reference_counter = cnt;
		p_prev = 0;
		p_next = 0;
		name[0] = 0;
		type = UNKNOWN;
	}

	virtual ~c_variable() {
		reject();
	}

	void reject( void ) {
		if( p_prev ) {
			p_prev->p_next = p_next;
		}
		if( p_next ) {
			p_next->p_prev = p_prev;
		}
	}

	c_variable *get_next( void ) {
		return p_next;
	}
	
	c_variable *get_prev( void ) {
		return p_prev;
	}

	const char *get_name( void ) const {
		return name;
	}

	int get_type_id( void ) const {
		return type;
	}

	void add_ref( void ) {
		reference_counter++;
		dlog.printf( "  Variable [%s] add reference (%d)\n", name, reference_counter );
	}
	
	void release_ref( void ) {
		dlog.printf( "  Variable [%s] release reference (%d)\n", name, reference_counter );
		reference_counter--;
		if( reference_counter == 0 ) {
			dlog.printf( "    Variable [%s] delete\n", name );
			delete this;
			return;
		}
	}

	virtual void clear( void ) {
	}

	void insert_prev( c_variable *p_new_prev ) {

		p_new_prev->reject();
		p_new_prev->p_next = this;
		p_new_prev->p_prev = p_prev;
		p_prev = p_new_prev;
	}

	void insert_next( c_variable *p_new_next ) {

		p_new_next->reject();
		p_new_next->p_prev = this;
		p_new_next->p_next = p_next;
		p_next = p_new_next;
	}

	void set_name( const char *p ) {
		std::strcpy( name, p );
	}
};

/* --------------------------------------------------------- */
class c_variable_label: public c_variable {
private:
	int value;
	c_program_line *p_line;

public:
	c_variable_label( void ): c_variable() {
		type = c_parser::PC_LABEL;
		value = 0;
		p_line = 0;
	}

	~c_variable_label() {
	}

	void set( int index ) {
		value = index;
	}

	void set_line( c_program_line *p ) {
		p_line = p;
	}

	int get( void ) const {
		return value;
	}
	c_program_line *get_line( void ) {
		return p_line;
	}
};

/* --------------------------------------------------------- */
class c_variable_integer: public c_variable {
private:
	int value;

public:
	c_variable_integer( void ): c_variable() {
		type = c_parser::PC_INUMBER;
		value = 0;
	}

	~c_variable_integer() {
	}

	void set( int v ) {
		value = v;
	}

	int get( void ) const {
		return value;
	}
};

/* --------------------------------------------------------- */
class c_variable_real: public c_variable {
private:
	double value;

public:
	c_variable_real( void ): c_variable() {
		type = c_parser::PC_RNUMBER;
		value = 0.;
	}

	~c_variable_real() {
	}

	void set( double v ) {
		value = v;
	}

	double get( void ) const {
		return value;
	}
};

/* --------------------------------------------------------- */
class c_variable_string: public c_variable {
private:
	char value[ 256 ];	/* WARNING: This is NOT null terminated string. */
	int length;

public:
	c_variable_string( void ): c_variable() {
		type = c_parser::PC_STRING;
		length = 0;
	}

	~c_variable_string() {
	}

	void set( const char *p, int l ) {
		length = l;
		std::memcpy( value, p, length );
	}

	const char *get( void ) const {
		return value;
	}

	int get_length( void ) const {
		return length;
	}
};

/* --------------------------------------------------------- */
class c_variable_integer_array: public c_variable {
private:
	int*	p_value;
	int		dimension;
	int		counts[ MAX_ARRAY_DIMENTION ];

public:
	c_variable_integer_array( void ): c_variable() {
		type = c_parser::PC_IARRAY;
		p_value = 0;
		dimension = 0;
	}

	~c_variable_integer_array() {
		delete[] p_value;
	}

	void clear( void );
	int set( int num_elements, int *p_element, int v );
	int dim( int num_elements, int *p_element );
	int get( int num_elements, int *p_element, int &v );
};

/* --------------------------------------------------------- */
class c_variable_real_array: public c_variable {
private:
	double*	p_value;
	int		dimension;
	int		counts[ MAX_ARRAY_DIMENTION ];

public:
	c_variable_real_array( void ): c_variable() {
		type = c_parser::PC_RARRAY;
		p_value = 0;
		dimension = 0;
	}
	~c_variable_real_array() {
		delete[] p_value;
	}

	void clear( void );
	int set( int num_elements, int *p_element, double v );
	int dim( int num_elements, int *p_element );
	int get( int num_elements, int *p_element, double &v );
};

/* --------------------------------------------------------- */
class c_string_element {
public:
	char value[ 256 ];	/* WARNING: This is NOT null terminated string. */
	int length;

	c_string_element( void ) {
		length = 0;
	}
};

/* --------------------------------------------------------- */
class c_variable_string_array: public c_variable {
private:
	c_string_element*	p_value;
	int		dimension;
	int		counts[ MAX_ARRAY_DIMENTION ];

public:
	c_variable_string_array( void ): c_variable() {
		type = c_parser::PC_RARRAY;
		p_value = 0;
		dimension = 0;
	}

	~c_variable_string_array() {
		delete[] p_value;
	}

	void clear( void );
	int set( int num_elements, int *p_element, const char *p_string, int length );
	int dim( int num_elements, int *p_element );
	int get( int num_elements, int *p_element, const char* &p_string, int &length );
};

/* --------------------------------------------------------- */
class c_variable_manager {
private:
	c_variable single_list;
	c_variable array_list;
	c_variable label_list;

	c_variable *search( c_variable *p_list, const char *p_name );

public:
	c_variable_manager( void ) {
	}

	template<class  c_variable_sub>
	c_variable* add_single( const char *p_name ) {
		c_variable *p;

		p = search( single_list.get_next(), p_name );
		if( p == 0 ) {
			/* new member */
			p = new c_variable_sub;
			p->set_name( p_name );
			single_list.insert_next( p );
		}
		else {
			/* update */
			p->add_ref();
		}
		return p;
	}

	template<class  c_variable_sub>
	c_variable* add_array( const char *p_name ) {
		c_variable *p;

		p = search( array_list.get_next(), p_name );
		if( p == 0 ) {
			/* new member */
			p = new c_variable_sub;
			p->set_name( p_name );
			array_list.insert_next( p );
		}
		else {
			/* update */
			p->add_ref();
		}
		return p;
	}

	c_variable_label* add_label( const char *p_name ) {
		c_variable_label *p;

		p = reinterpret_cast<c_variable_label*>( search( label_list.get_next(), p_name ) );
		if( p == 0 ) {
			/* new member */
			p = new c_variable_label;
			p->set_name( p_name );
			label_list.insert_next( p );
		}
		else {
			/* update */
			p->add_ref();
		}
		return p;
	}

	void line_release( const unsigned char *p_precompiled_code, int length );
	void array_clear( void );
};

#endif
